import java.awt.Graphics;
import java.awt.Color;

/**
 * Une pice du jeu est compose d'un certain nombre de blocs, et d'une couleur.
 * Le positionnement de chacun des blocs par rapport aux autres est aussi trs
 * important. Il existe 7 pieces dans le jeu original. Chacune est compose de
 * quatre blocs. Voici ces 7 pices, le bloc central autour duquel les autres
 * tournent est reprsent par un X majuscule. Le numro de la pice, tel qu'il
 * est utilise plus bas, est indiqu.
 *	<pre>
 *  x            x          x x    x x
 *  x X x    x X x        x X        X x
 *    0        1            2        3
 *
 *    x      x x
 *  x X x    X x          x X x x
 *
 *    4        5            6
 * </pre>
 */

class Piece {
	
	// Champs de classe
	public static final int ROTATE = 1;
	public static final int LEFT = 2;
	public static final int DOWN = 3;
	public static final int RIGHT = 4;
	public static final int DROP = 5;
	
	/**
	 * Tableau tri-dimensionnelle d'entiers, dfinissant les diffrentes pices
	 * du jeu. Le premier indice est le numro d'identification de la pice, le
	 * deuxime indice est le numro du bloc et le troisime est le numro de la
	 * donne (0 pour abscisse et 1 pour ordonne). Les pices peuvent avoir un
	 * nombre de blocs diffrents les unes par rapport aux autres. */
	private static final int[][][] MODELS = {
		{ {0, 0},{-1, -1},{-1, 0},{1, 0} },
		{ {0, 0},{-1, 0},{1, 0},{1, -1} },
		{ {0, 0},{-1, 0},{0, -1},{1, -1} },
		{ {0, 0},{-1, -1},{0, -1},{1, 0} },
		{ {0, 0},{-1, 0},{0, -1},{1, 0} },
		{ {0, 0},{0, -1},{1, 0},{1, -1} },
		{ {0, 0},{0, -2},{0, -1},{0, 1} },
		// Ci-dessous, les pices ajoutes au jeu original
		{ {0, 0},{0, -1},{1, 0} },
		{ {0, 0},{-1, -1},{-1, 0},{1, 0},{1,-1} },
		{ {0, 0}}
	};
	/** True si on joue en mode "monochrome". */
	private static boolean monochrome;
	/** Couleur de toutes les pices si on joue en monochrome. @see Tetris.monochrome */
	private static Color uniColor;
	/** Couleurs des pices lorsque le jeu n'est pas monochrome. @see Tetris.monochrome */
	private static Color[] colors = new Color[MODELS.length];
	/**
	 * Liste des pices utiliss dans le jeu. Les lments du tableau sont des indices
	 * de MODELS. Ainsi, models[i] = i pour i = 0...6 signifie qu'on va jouer avec les
	 * sept premires pices de MODELS. @see Piece.MODELS */
	private static int[] models;
	
	// Champs de variable
	/** Couleur de la pice. */
	private Color color;
	/** Tableau des blocs composant la pice. */
	private Block[] blocks;
	

	// Constructeurs
	/** Cre une pice selon MODELS[n] puis la dplace de x colonnes et de y lignes. */
	public Piece(int n, int x, int y) {
		color = (monochrome)?uniColor:colors[n];
		blocks = new Block[MODELS[n].length];
		for(int i=0;i<blocks.length;i++) {
			blocks[i] = new Block(MODELS[n][i][0] + x, MODELS[n][i][1] + y);
		}
	}
	public Piece(int x, int y){this(getRandomModelNumber(),x,y);}
	public Piece(){this(getRandomModelNumber(),0,0);}


	// Mthodes
	/** Dplace la pice en fixant son bloc central sur (horizontal, vertical). */
	public void go(int horizontal, int vertical) {
			int x = blocks[0].getX();
			int y = blocks[0].getY();
			for(int i=0;i<blocks.length;i++)
				blocks[i].move(horizontal-x, vertical-y);
	}
	/** Dplace une pice de horizontal colonnes et de vertical lignes. */
	public void move(int horizontal, int vertical) {
		for(int i=0;i<blocks.length;i++)
	  	blocks[i].move(horizontal, vertical);
	}
	/** Dplace une pice en se servant des constantes ROTATE, LEFT, DOWN et RIGHT. */
	public void move(int movement) {
		switch(movement) {
			case ROTATE :
				rotate(true);
				break;
			case -ROTATE :
				rotate(false);
				break;
			case DOWN :
				move(0,-1);
				break;
			case -DOWN :
				move(0,1);
				break;
			case LEFT :	case -RIGHT :
				move(-1,0);
				break;
			case RIGHT : case -LEFT :
				move(1,0);
				break;
		}
	}
	
	/**
	 * Faire tourner une piece autour du bloc central :
	 * si a parait un peu compliqu au dbut, un petit
	 * algorithme qui marche bien revient  exprimer les
	 * coordonnes de chaque bloc relativement  celles
	 * du bloc central, puis d'changer les valeurs des
	 * abscisses et des ordonnes en multipliant l'une
	 * des deux par (-1). Ensuite il suffit de repasser
	 * les coordonnes en "absolue" en ajoutant celles
	 * du bloc central. Mathmatiquement :
	 * <pre>
	 * Avec :
	 *
	 * (x0,y0)	le bloc central
	 * (x,y)		le bloc  dplacer
	 *
	 * On fait :
	 *
	 * x = (y-y0)+x0
	 * y = -(x-x0)+y0		pour tourner dans le sens des aiguilles d'une montre
	 * 
 	 * x = -(y-y0)+x0
	 * y = (x-x0)+y0		pour tourner dans le sens contraire
	 *
	 * </pre>
	 * Attention : a n'apparait pas ci-dessus par souci
	 * de clart mais on a besoin de la valeur "initiale"
	 * de x pour calculer la nouvelle valeur de y.
	 *
	 * @param clockwise Mettre true si la rotation doit se fait dans
	 * le sens des aiguilles d'une montre.
	 */
	public void rotate(boolean clockwise) {
		int temp0x = blocks[0].getX();
	  int temp0y = blocks[0].getY();
	  for(int i=1;i<blocks.length;i++) {
		  int tempix = blocks[i].getX();
			int tempiy = blocks[i].getY();
		  if (clockwise == true)	{	//	tourner dans le sens des aiguilles
				blocks[i].setX(tempiy - temp0y + temp0x);
				blocks[i].setY(- tempix + temp0x + temp0y);
		  } else {									//	tourner dans l'autre sens
				blocks[i].setX(- tempiy + temp0y + temp0x);
				blocks[i].setY(tempix - temp0x + temp0y);
		  }
	  }
	}
	
	/**
	 * Afficher la pice dans le Graphics g avec la couleur c.
	 * On met outPainting  false si on veut n'afficher que les blocs
	 * dont l'ordonne ne dpasse Grid.rows. En pratique, on met false
	 * pour afficher la pice qui est entrain de jouer et true pour
	 * afficher la pice  venir. @see Grid.ROWS */  
  public void paint(Graphics g, Color c, boolean outPainting) {
		g.setColor(c);
		for(int i=0;i<blocks.length;i++)
		  if (blocks[i].getY() < Grid.ROWS || outPainting)
			  blocks[i].paint(g);
	}
	/** Afficher la pice dans sa couleur. */
	public void paint(Graphics g, boolean b) {
		paint(g,color,b);
	}
	/** Idem que ci-dessus en n'affichant que le bloc dont l'ordonne est infrieure  Grid.ROWS. */
	public void paint(Graphics g) {
	  paint(g,color,false);
  }


	// Accessors
	/**
	 * Il ne s'agit pas d'un vritable "accessor" mais il est pratique de le considrer comme tel.
	 * Retourne un lment de models tir alatoirement. @return Un entier entre 0 et models.length. */
	private static int getRandomModelNumber() {return models[(int)(Math.random()*models.length)];}
	/**
	 * Il ne s'agit pas vraiment d'un "accessor" mais il est pratique de considrer comme tel.
	 * Retourne alatoirement une couleur parmi celles des pices utilises par le jeu.
	 * @return La couleur d'une des pices utilises dans le jeu. */
	public static Color getRandomUsedColor() {return (monochrome)?uniColor:colors[getRandomModelNumber()];}
	public static Color getUniColor() {return uniColor;}
	public static Color[] getColors() {return colors;}
	public static boolean isMonochrome() {return monochrome;}
	public Block[] getBlocks() {return blocks;}
	public Color getColor() {return color;}

			
	// Mutators
	public static void setMonochrome(boolean b) {monochrome = b;}
	public static void setUniColor(Color c) {uniColor = c;}
	public static void setColors(int i, Color c) {colors[i] = c;}
	/**
	 * Cette mthode static, il faut la lancer en dbut de partie.
	 * Elle rempli models avec les indices des MODELS correspondant aux
	 * pices que l'on souhaite utilis. Pour ce faire, elle utilise colors en
	 * considrant que pice d'indice i est non dsire si colors[i] = null. */
	public static void setModels() {
		int i=0,j=0;
		for(;i<colors.length;i++)
			if(colors[i] != null)
				j++;
		models = new int[j];
		for(i=0,j=0;i<models.length;i++,j++) {
			while(colors[j] == null)
				j++;
			models[i] = j;
		}
	}


}